from .demo_func import DemoFunc
from .demo_func import DemoFuncAckley
from .demo_func import DemoFuncGrienwank
from .demo_func import DemoFuncMichalewicz
from .demo_func import DemoFuncPiston
from .demo_func import DemoFuncRastrigin
from .demo_func import DemoFuncRosenbrock
from .demo_func import DemoFuncSchwefel
